import typing

import py_gen_ml as pgml

from . import unit_base as base
from . import unit_patch as patch

EnumSweepField = typing.Union[
    pgml.Choice[base.Enum],
    typing.Literal['any'],
    base.Enum,
]


class Int32TestSweep(pgml.Sweeper[patch.Int32TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Int32TestSweepField = typing.Union[
    Int32TestSweep,
    pgml.NestedChoice[Int32TestSweep, patch.Int32TestPatch],  # type: ignore
]


class Int64TestSweep(pgml.Sweeper[patch.Int64TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Int64TestSweepField = typing.Union[
    Int64TestSweep,
    pgml.NestedChoice[Int64TestSweep, patch.Int64TestPatch],  # type: ignore
]


class Uint32TestSweep(pgml.Sweeper[patch.Uint32TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Uint32TestSweepField = typing.Union[
    Uint32TestSweep,
    pgml.NestedChoice[Uint32TestSweep, patch.Uint32TestPatch],  # type: ignore
]


class Uint64TestSweep(pgml.Sweeper[patch.Uint64TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Uint64TestSweepField = typing.Union[
    Uint64TestSweep,
    pgml.NestedChoice[Uint64TestSweep, patch.Uint64TestPatch],  # type: ignore
]


class Sint32TestSweep(pgml.Sweeper[patch.Sint32TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Sint32TestSweepField = typing.Union[
    Sint32TestSweep,
    pgml.NestedChoice[Sint32TestSweep, patch.Sint32TestPatch],  # type: ignore
]


class Sint64TestSweep(pgml.Sweeper[patch.Sint64TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Sint64TestSweepField = typing.Union[
    Sint64TestSweep,
    pgml.NestedChoice[Sint64TestSweep, patch.Sint64TestPatch],  # type: ignore
]


class Fixed32TestSweep(pgml.Sweeper[patch.Fixed32TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Fixed32TestSweepField = typing.Union[
    Fixed32TestSweep,
    pgml.NestedChoice[Fixed32TestSweep, patch.Fixed32TestPatch],  # type: ignore
]


class Fixed64TestSweep(pgml.Sweeper[patch.Fixed64TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Fixed64TestSweepField = typing.Union[
    Fixed64TestSweep,
    pgml.NestedChoice[Fixed64TestSweep, patch.Fixed64TestPatch],  # type: ignore
]


class Sfixed32TestSweep(pgml.Sweeper[patch.Sfixed32TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Sfixed32TestSweepField = typing.Union[
    Sfixed32TestSweep,
    pgml.NestedChoice[Sfixed32TestSweep, patch.Sfixed32TestPatch],  # type: ignore
]


class Sfixed64TestSweep(pgml.Sweeper[patch.Sfixed64TestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Sfixed64TestSweepField = typing.Union[
    Sfixed64TestSweep,
    pgml.NestedChoice[Sfixed64TestSweep, patch.Sfixed64TestPatch],  # type: ignore
]


class BoolTestSweep(pgml.Sweeper[patch.BoolTestPatch]):
    value: typing.Optional[pgml.BoolSweep] = None


BoolTestSweepField = typing.Union[
    BoolTestSweep,
    pgml.NestedChoice[BoolTestSweep, patch.BoolTestPatch],  # type: ignore
]


class FloatTestSweep(pgml.Sweeper[patch.FloatTestPatch]):
    value: typing.Optional[pgml.FloatSweep] = None


FloatTestSweepField = typing.Union[
    FloatTestSweep,
    pgml.NestedChoice[FloatTestSweep, patch.FloatTestPatch],  # type: ignore
]


class DoubleTestSweep(pgml.Sweeper[patch.DoubleTestPatch]):
    value: typing.Optional[pgml.FloatSweep] = None


DoubleTestSweepField = typing.Union[
    DoubleTestSweep,
    pgml.NestedChoice[DoubleTestSweep, patch.DoubleTestPatch],  # type: ignore
]


class StringTestSweep(pgml.Sweeper[patch.StringTestPatch]):
    value: typing.Optional[pgml.StrSweep] = None


StringTestSweepField = typing.Union[
    StringTestSweep,
    pgml.NestedChoice[StringTestSweep, patch.StringTestPatch],  # type: ignore
]


class BytesTestSweep(pgml.Sweeper[patch.BytesTestPatch]):
    value: typing.Optional[pgml.BytesSweep] = None


BytesTestSweepField = typing.Union[
    BytesTestSweep,
    pgml.NestedChoice[BytesTestSweep, patch.BytesTestPatch],  # type: ignore
]


class EnumTestSweep(pgml.Sweeper[patch.EnumTestPatch]):
    value: typing.Optional[EnumSweepField] = None


EnumTestSweepField = typing.Union[
    EnumTestSweep,
    pgml.NestedChoice[EnumTestSweep, patch.EnumTestPatch],  # type: ignore
]


class EnumDefaultTestSweep(pgml.Sweeper[patch.EnumDefaultTestPatch]):
    value: typing.Optional[EnumSweepField] = None


EnumDefaultTestSweepField = typing.Union[
    EnumDefaultTestSweep,
    pgml.NestedChoice[EnumDefaultTestSweep, patch.EnumDefaultTestPatch],  # type: ignore
]


class OneofTestSweep(pgml.Sweeper[patch.OneofTestPatch]):
    value: typing.Optional[typing.Union[pgml.Sweeper[pgml.IntSweep], pgml.Sweeper[pgml.StrSweep]]] = None


OneofTestSweepField = typing.Union[
    OneofTestSweep,
    pgml.NestedChoice[OneofTestSweep, patch.OneofTestPatch],  # type: ignore
]


class RepeatedTestSweep(pgml.Sweeper[patch.RepeatedTestPatch]):
    values: typing.Optional[pgml.IntSweep] = None


RepeatedTestSweepField = typing.Union[
    RepeatedTestSweep,
    pgml.NestedChoice[RepeatedTestSweep, patch.RepeatedTestPatch],  # type: ignore
]


class OptionalTestSweep(pgml.Sweeper[patch.OptionalTestPatch]):
    _value: typing.Optional[typing.Union[pgml.Sweeper[pgml.IntSweep]]] = None


OptionalTestSweepField = typing.Union[
    OptionalTestSweep,
    pgml.NestedChoice[OptionalTestSweep, patch.OptionalTestPatch],  # type: ignore
]


class Int32DefaultTestSweep(pgml.Sweeper[patch.Int32DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Int32DefaultTestSweepField = typing.Union[
    Int32DefaultTestSweep,
    pgml.NestedChoice[Int32DefaultTestSweep, patch.Int32DefaultTestPatch],  # type: ignore
]


class Int64DefaultTestSweep(pgml.Sweeper[patch.Int64DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Int64DefaultTestSweepField = typing.Union[
    Int64DefaultTestSweep,
    pgml.NestedChoice[Int64DefaultTestSweep, patch.Int64DefaultTestPatch],  # type: ignore
]


class Uint32DefaultTestSweep(pgml.Sweeper[patch.Uint32DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Uint32DefaultTestSweepField = typing.Union[
    Uint32DefaultTestSweep,
    pgml.NestedChoice[Uint32DefaultTestSweep, patch.Uint32DefaultTestPatch],  # type: ignore
]


class Uint64DefaultTestSweep(pgml.Sweeper[patch.Uint64DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Uint64DefaultTestSweepField = typing.Union[
    Uint64DefaultTestSweep,
    pgml.NestedChoice[Uint64DefaultTestSweep, patch.Uint64DefaultTestPatch],  # type: ignore
]


class Sint32DefaultTestSweep(pgml.Sweeper[patch.Sint32DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Sint32DefaultTestSweepField = typing.Union[
    Sint32DefaultTestSweep,
    pgml.NestedChoice[Sint32DefaultTestSweep, patch.Sint32DefaultTestPatch],  # type: ignore
]


class Sint64DefaultTestSweep(pgml.Sweeper[patch.Sint64DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Sint64DefaultTestSweepField = typing.Union[
    Sint64DefaultTestSweep,
    pgml.NestedChoice[Sint64DefaultTestSweep, patch.Sint64DefaultTestPatch],  # type: ignore
]


class Fixed32DefaultTestSweep(pgml.Sweeper[patch.Fixed32DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Fixed32DefaultTestSweepField = typing.Union[
    Fixed32DefaultTestSweep,
    pgml.NestedChoice[Fixed32DefaultTestSweep, patch.Fixed32DefaultTestPatch],  # type: ignore
]


class Fixed64DefaultTestSweep(pgml.Sweeper[patch.Fixed64DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Fixed64DefaultTestSweepField = typing.Union[
    Fixed64DefaultTestSweep,
    pgml.NestedChoice[Fixed64DefaultTestSweep, patch.Fixed64DefaultTestPatch],  # type: ignore
]


class Sfixed32DefaultTestSweep(pgml.Sweeper[patch.Sfixed32DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Sfixed32DefaultTestSweepField = typing.Union[
    Sfixed32DefaultTestSweep,
    pgml.NestedChoice[Sfixed32DefaultTestSweep, patch.Sfixed32DefaultTestPatch],  # type: ignore
]


class Sfixed64DefaultTestSweep(pgml.Sweeper[patch.Sfixed64DefaultTestPatch]):
    value: typing.Optional[pgml.IntSweep] = None


Sfixed64DefaultTestSweepField = typing.Union[
    Sfixed64DefaultTestSweep,
    pgml.NestedChoice[Sfixed64DefaultTestSweep, patch.Sfixed64DefaultTestPatch],  # type: ignore
]


class BoolDefaultTestSweep(pgml.Sweeper[patch.BoolDefaultTestPatch]):
    value: typing.Optional[pgml.BoolSweep] = None


BoolDefaultTestSweepField = typing.Union[
    BoolDefaultTestSweep,
    pgml.NestedChoice[BoolDefaultTestSweep, patch.BoolDefaultTestPatch],  # type: ignore
]


class FloatDefaultTestSweep(pgml.Sweeper[patch.FloatDefaultTestPatch]):
    value: typing.Optional[pgml.FloatSweep] = None


FloatDefaultTestSweepField = typing.Union[
    FloatDefaultTestSweep,
    pgml.NestedChoice[FloatDefaultTestSweep, patch.FloatDefaultTestPatch],  # type: ignore
]


class DoubleDefaultTestSweep(pgml.Sweeper[patch.DoubleDefaultTestPatch]):
    value: typing.Optional[pgml.FloatSweep] = None


DoubleDefaultTestSweepField = typing.Union[
    DoubleDefaultTestSweep,
    pgml.NestedChoice[DoubleDefaultTestSweep, patch.DoubleDefaultTestPatch],  # type: ignore
]


class StringDefaultTestSweep(pgml.Sweeper[patch.StringDefaultTestPatch]):
    value: typing.Optional[pgml.StrSweep] = None


StringDefaultTestSweepField = typing.Union[
    StringDefaultTestSweep,
    pgml.NestedChoice[StringDefaultTestSweep, patch.StringDefaultTestPatch],  # type: ignore
]


class BytesDefaultTestSweep(pgml.Sweeper[patch.BytesDefaultTestPatch]):
    value: typing.Optional[pgml.BytesSweep] = None


BytesDefaultTestSweepField = typing.Union[
    BytesDefaultTestSweep,
    pgml.NestedChoice[BytesDefaultTestSweep, patch.BytesDefaultTestPatch],  # type: ignore
]


class OneofDefaultTestSweep(pgml.Sweeper[patch.OneofDefaultTestPatch]):
    value: typing.Optional[typing.Union[pgml.Sweeper[pgml.IntSweep], pgml.Sweeper[pgml.StrSweep]]] = None


OneofDefaultTestSweepField = typing.Union[
    OneofDefaultTestSweep,
    pgml.NestedChoice[OneofDefaultTestSweep, patch.OneofDefaultTestPatch],  # type: ignore
]


class OptionalDefaultTestSweep(pgml.Sweeper[patch.OptionalDefaultTestPatch]):
    _value: typing.Optional[typing.Union[pgml.Sweeper[pgml.IntSweep]]] = None


OptionalDefaultTestSweepField = typing.Union[
    OptionalDefaultTestSweep,
    pgml.NestedChoice[OptionalDefaultTestSweep, patch.OptionalDefaultTestPatch],  # type: ignore
]


class ExplicitCLIArgTestSweep(pgml.Sweeper[patch.ExplicitCLIArgTestPatch]):
    bar: typing.Optional[pgml.StrSweep] = None


ExplicitCLIArgTestSweepField = typing.Union[
    ExplicitCLIArgTestSweep,
    pgml.NestedChoice[ExplicitCLIArgTestSweep, patch.ExplicitCLIArgTestPatch],  # type: ignore
]


class ImplicitCLIArgTestSweep(pgml.Sweeper[patch.ImplicitCLIArgTestPatch]):
    bar: typing.Optional[pgml.StrSweep] = None


ImplicitCLIArgTestSweepField = typing.Union[
    ImplicitCLIArgTestSweep,
    pgml.NestedChoice[ImplicitCLIArgTestSweep, patch.ImplicitCLIArgTestPatch],  # type: ignore
]


class NestedFooTestSweep(pgml.Sweeper[patch.NestedFooTestPatch]):
    foo: typing.Optional[pgml.StrSweep] = None


NestedFooTestSweepField = typing.Union[
    NestedFooTestSweep,
    pgml.NestedChoice[NestedFooTestSweep, patch.NestedFooTestPatch],  # type: ignore
]


class NestedBarTestSweep(pgml.Sweeper[patch.NestedBarTestPatch]):
    foo_0: typing.Optional[NestedFooTestSweepField] = None
    foo_1: typing.Optional[NestedFooTestSweepField] = None


NestedBarTestSweepField = typing.Union[
    NestedBarTestSweep,
    pgml.NestedChoice[NestedBarTestSweep, patch.NestedBarTestPatch],  # type: ignore
]


class RepeatedNestedBarTestSweep(pgml.Sweeper[patch.RepeatedNestedBarTestPatch]):
    bar: typing.Optional[NestedBarTestSweepField] = None


RepeatedNestedBarTestSweepField = typing.Union[
    RepeatedNestedBarTestSweep,
    pgml.NestedChoice[RepeatedNestedBarTestSweep, patch.RepeatedNestedBarTestPatch],  # type: ignore
]



import io
import sys
from typing import Dict

import google.protobuf.compiler.plugin_pb2
import google.protobuf.json_format
import protogen

from py_gen_ml.logging.setup_logger import setup_logger
from py_gen_ml.plugin.base_model_generator import BaseModelGenerator
from py_gen_ml.plugin.cli_args_generator import CliArgsGenerator
from py_gen_ml.plugin.generator import GenTask, InitGenerator
from py_gen_ml.plugin.json_schema_task_generator import JsonSchemaTaskGenerator
from py_gen_ml.plugin.sweep_model_generator import SweepModelGenerator

logger = setup_logger(__name__)


class _Plugin:

    def __init__(self) -> None:
        self._gen_tasks: list[GenTask] = []

    def generate(self, plugin: protogen.Plugin) -> None:
        """
        Generate the necessary files for the plugin.

        This function generates the base model, patch model, sweep model, and CLI arguments
        for the given plugin. It uses the BaseModelGenerator, SweepModelGenerator, and
        CliArgsGenerator classes to create the respective files.
        """
        for generator in [
            BaseModelGenerator(plugin, is_patch=False, suffix='_base.py'),
            BaseModelGenerator(plugin, is_patch=True, suffix='_patch.py'),
            SweepModelGenerator(plugin, suffix='_sweep.py'),
            CliArgsGenerator(plugin, suffix='_cli_args.py'),
            InitGenerator(plugin),
        ]:
            generator.generate_code()
            self._gen_tasks.extend(generator.json_schema_gen_tasks)

        JsonSchemaTaskGenerator(plugin, self._gen_tasks).generate_code()

    @property
    def json_schema_gen_tasks(self) -> list[GenTask]:
        return self._gen_tasks


def run() -> None:
    """
    Run the plugin to generate the necessary files.

    This function sets up the plugin options and runs the generate function to create
    the required files. It uses the protogen.Options class to configure the supported
    features and then executes the generate function.
    """
    logger.debug('Running py-gen-ml plugin')

    input_stream = sys.stdin.buffer
    request = google.protobuf.compiler.plugin_pb2.CodeGeneratorRequest()
    request.ParseFromString(input_stream.read())

    if request.HasField('parameter') and request.parameter:
        parameter: Dict[str, str] = {}
        for param in request.parameter.split(','):
            if param == '':
                # Ignore empty parameters.
                continue
            splits = param.split('=', 1)  # maximum one split
            if len(splits) == 1:
                k, v = splits[0], ''
            else:
                k, v = splits
            parameter[k] = v
        if parameter.get('export_code_generator_request', 'false').lower() == 'true':
            with open('request.pbbin', 'wb') as f:
                f.write(request.SerializeToString())

    input_stream = io.BytesIO(request.SerializeToString())
    opts = protogen.Options(
        input=input_stream,
        supported_features=[
            google.protobuf.compiler.plugin_pb2.CodeGeneratorResponse.Feature.FEATURE_PROTO3_OPTIONAL,  # type: ignore
        ],
    )

    input_stream.seek(0)
    plugin = _Plugin()
    opts.run(plugin.generate)



import pathlib

from .pgml_out_test import unit_base


def test_int32() -> None:
    assert unit_base.Int32Test(value=1).value == 1


def test_int64() -> None:
    assert unit_base.Int64Test(value=1).value == 1


def test_uint32() -> None:
    assert unit_base.Uint32Test(value=1).value == 1


def test_uint64() -> None:
    assert unit_base.Uint64Test(value=1).value == 1


def test_sint32() -> None:
    assert unit_base.Sint32Test(value=1).value == 1


def test_sint64() -> None:
    assert unit_base.Sint64Test(value=1).value == 1


def test_fixed32() -> None:
    assert unit_base.Fixed32Test(value=1).value == 1


def test_fixed64() -> None:
    assert unit_base.Fixed64Test(value=1).value == 1


def test_sfixed32() -> None:
    assert unit_base.Sfixed32Test(value=1).value == 1


def test_sfixed64() -> None:
    assert unit_base.Sfixed64Test(value=1).value == 1


def test_float() -> None:
    assert unit_base.FloatTest(value=1).value == 1


def test_double() -> None:
    assert unit_base.DoubleTest(value=1).value == 1


def test_bool() -> None:
    assert unit_base.BoolTest(value=True).value == True


def test_string() -> None:
    assert unit_base.StringTest(value='test').value == 'test'


def test_bytes() -> None:
    assert unit_base.BytesTest(value=b'test').value == b'test'


def test_enum() -> None:
    assert unit_base.EnumTest(value=unit_base.Enum.VALUE_1).value == unit_base.Enum.VALUE_1


def test_int32_default() -> None:
    assert unit_base.Int32DefaultTest().value == 1


def test_int64_default() -> None:
    assert unit_base.Int64DefaultTest().value == 1


def test_uint32_default() -> None:
    assert unit_base.Uint32DefaultTest().value == 1


def test_uint64_default() -> None:
    assert unit_base.Uint64DefaultTest().value == 1


def test_sint32_default() -> None:
    assert unit_base.Sint32DefaultTest().value == 1


def test_sint64_default() -> None:
    assert unit_base.Sint64DefaultTest().value == 1


def test_fixed32_default() -> None:
    assert unit_base.Fixed32DefaultTest().value == 1


def test_fixed64_default() -> None:
    assert unit_base.Fixed64DefaultTest().value == 1


def test_sfixed32_default() -> None:
    assert unit_base.Sfixed32DefaultTest().value == 1


def test_sfixed64_default() -> None:
    assert unit_base.Sfixed64DefaultTest().value == 1


def test_float_default() -> None:
    assert unit_base.FloatDefaultTest().value == 1


def test_double_default() -> None:
    assert unit_base.DoubleDefaultTest().value == 1


def test_bool_default() -> None:
    assert unit_base.BoolDefaultTest().value == True


def test_string_default() -> None:
    assert unit_base.StringDefaultTest().value == 'test'


def test_bytes_default() -> None:
    assert unit_base.BytesDefaultTest().value == b'test'


def test_enum_default() -> None:
    assert unit_base.EnumDefaultTest().value == unit_base.Enum.VALUE_1


def test_int32_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Int32Test.from_yaml_file(str(test_file)) == unit_base.Int32Test(value=1)


def test_int64_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Int64Test.from_yaml_file(str(test_file)) == unit_base.Int64Test(value=1)


def test_uint32_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Uint32Test.from_yaml_file(str(test_file)) == unit_base.Uint32Test(value=1)


def test_uint64_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Uint64Test.from_yaml_file(str(test_file)) == unit_base.Uint64Test(value=1)


def test_sint32_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Sint32Test.from_yaml_file(str(test_file)) == unit_base.Sint32Test(value=1)


def test_sint64_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Sint64Test.from_yaml_file(str(test_file)) == unit_base.Sint64Test(value=1)


def test_fixed32_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Fixed32Test.from_yaml_file(str(test_file)) == unit_base.Fixed32Test(value=1)


def test_fixed64_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Fixed64Test.from_yaml_file(str(test_file)) == unit_base.Fixed64Test(value=1)


def test_sfixed32_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Sfixed32Test.from_yaml_file(str(test_file)) == unit_base.Sfixed32Test(value=1)


def test_sfixed64_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.Sfixed64Test.from_yaml_file(str(test_file)) == unit_base.Sfixed64Test(value=1)


def test_float_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.FloatTest.from_yaml_file(str(test_file)) == unit_base.FloatTest(value=1)


def test_double_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: 1')
    assert unit_base.DoubleTest.from_yaml_file(str(test_file)) == unit_base.DoubleTest(value=1)


def test_bool_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: true')
    assert unit_base.BoolTest.from_yaml_file(str(test_file)) == unit_base.BoolTest(value=True)


def test_string_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: test')
    assert unit_base.StringTest.from_yaml_file(str(test_file)) == unit_base.StringTest(value='test')


def test_bytes_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: test')
    assert unit_base.BytesTest.from_yaml_file(str(test_file)) == unit_base.BytesTest(value=b'test')


def test_enum_yaml(tmp_path: pathlib.Path) -> None:
    test_file = tmp_path / 'test.yaml'
    test_file.write_text('value: VALUE_1')
    assert unit_base.EnumTest.from_yaml_file(str(test_file)) == unit_base.EnumTest(value=unit_base.Enum.VALUE_1)



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml

from . import mlp_base as base


class MLPParsingDemoPatch(pgml.YamlBaseModel):
    """MLP is a simple multi-layer perceptron."""

    num_layers: typing.Optional[int] = None
    """Number of layers in the MLP."""

    num_units: typing.Optional[int] = None
    """Number of units in each layer."""

    activation: typing.Optional[base.Activation] = None
    """Activation function to use."""



import typing

import py_gen_ml as pgml

from . import dataloader_patch as patch
from . import dataloader_base as base


class DataLoaderConfigSweep(pgml.Sweeper[patch.DataLoaderConfigPatch]):
    """DataLoader configuration"""

    batch_size: typing.Optional[pgml.IntSweep] = None
    """Batch size"""

    num_workers: typing.Optional[pgml.IntSweep] = None
    """Number of workers"""

    pin_memory: typing.Optional[pgml.BoolSweep] = None
    """Pin memory"""

    persistent_workers: typing.Optional[pgml.BoolSweep] = None
    """Persistent workers"""

    prefetch_factor: typing.Optional[pgml.IntSweep] = None
    """Prefetch factor"""



DataLoaderConfigSweepField = typing.Union[
    DataLoaderConfigSweep,
    pgml.NestedChoice[DataLoaderConfigSweep, patch.DataLoaderConfigPatch],  # type: ignore
]




import typing

import py_gen_ml as pgml

from . import cli_demo_deep_patch as patch
from . import cli_demo_deep_base as base


class ModelSweep(pgml.Sweeper[patch.ModelPatch]):
    """Model configuration"""

    num_layers: typing.Optional[pgml.IntSweep] = None
    """Number of layers"""


ModelSweepField = typing.Union[
    ModelSweep,
    pgml.NestedChoice[ModelSweep, patch.ModelPatch],  # type: ignore
]


class TrainingSweep(pgml.Sweeper[patch.TrainingPatch]):
    """Training configuration"""

    num_epochs: typing.Optional[pgml.IntSweep] = None
    """Number of epochs"""


TrainingSweepField = typing.Union[
    TrainingSweep,
    pgml.NestedChoice[TrainingSweep, patch.TrainingPatch],  # type: ignore
]


class DatasetSweep(pgml.Sweeper[patch.DatasetPatch]):
    """Dataset configuration"""

    path: typing.Optional[pgml.StrSweep] = None
    """Path to the dataset"""


DatasetSweepField = typing.Union[
    DatasetSweep,
    pgml.NestedChoice[DatasetSweep, patch.DatasetPatch],  # type: ignore
]


class DataSweep(pgml.Sweeper[patch.DataPatch]):
    """Data config"""

    train_dataset: typing.Optional[DatasetSweepField] = None
    """Path to the dataset"""

    test_dataset: typing.Optional[DatasetSweepField] = None
    """Path to the dataset"""

    num_workers: typing.Optional[pgml.IntSweep] = None
    """Number of workers for loading the dataset"""


DataSweepField = typing.Union[
    DataSweep,
    pgml.NestedChoice[DataSweep, patch.DataPatch],  # type: ignore
]


class CliDemoDeepSweep(pgml.Sweeper[patch.CliDemoDeepPatch]):
    """Global configuration"""

    data: typing.Optional[DataSweepField] = None
    """Dataset configuration"""

    model: typing.Optional[ModelSweepField] = None
    """Model configuration"""

    training: typing.Optional[TrainingSweepField] = None
    """Training configuration"""



CliDemoDeepSweepField = typing.Union[
    CliDemoDeepSweep,
    pgml.NestedChoice[CliDemoDeepSweep, patch.CliDemoDeepPatch],  # type: ignore
]




import typing

import py_gen_ml as pgml

from . import default_enum_patch as patch
from . import default_enum_base as base

ActivationSweepField = typing.Union[
    pgml.Choice[base.Activation],
    typing.Literal['any'],
    base.Activation,
]


class LinearSweep(pgml.Sweeper[patch.LinearPatch]):
    """Linear layer"""

    in_features: typing.Optional[pgml.IntSweep] = None
    """Number of input features"""

    out_features: typing.Optional[pgml.IntSweep] = None
    """Number of output features"""

    activation: typing.Optional[ActivationSweepField] = None
    """Activation function"""



LinearSweepField = typing.Union[
    LinearSweep,
    pgml.NestedChoice[LinearSweep, patch.LinearPatch],  # type: ignore
]




import typing

import py_gen_ml as pgml

from . import builder_demo_patch as patch
from . import builder_demo_base as base


class LinearSweep(pgml.Sweeper[patch.LinearPatch]):
    """Linear layer configuration"""

    in_features: typing.Optional[pgml.IntSweep] = None
    """Number of input features"""

    out_features: typing.Optional[pgml.IntSweep] = None
    """Number of output features"""

    bias: typing.Optional[pgml.BoolSweep] = None
    """Bias"""


LinearSweepField = typing.Union[
    LinearSweep,
    pgml.NestedChoice[LinearSweep, patch.LinearPatch],  # type: ignore
]


class MLPSweep(pgml.Sweeper[patch.MLPPatch]):
    """MLP configuration"""

    layers: typing.Optional[LinearSweepField] = None
    """Linear layers"""



MLPSweepField = typing.Union[
    MLPSweep,
    pgml.NestedChoice[MLPSweep, patch.MLPPatch],  # type: ignore
]




import os
import time
from functools import lru_cache
from typing import Any, List

import optuna
import torch
import torch.utils.data
import torchvision
import torchvision.transforms as transforms
import typer
from pgml_out.dataloader_base import DataLoaderConfig
from pgml_out.dataloader_sweep import DataLoaderConfigSweep

import py_gen_ml as pgml


def get_transform() -> transforms.Compose:
    return transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
    ])


app = typer.Typer(pretty_exceptions_enable=False)


@lru_cache
def get_dataset() -> torch.utils.data.Dataset[Any]:
    dataset = torchvision.datasets.CIFAR10(
        root=f"{os.environ['HOME']}/data/torchvision", train=True, download=True, transform=get_transform()
    )
    dataset = torch.utils.data.Subset(dataset, indices=list(range(10000)))
    return dataset


def dataloader_speed_benchmark(config: DataLoaderConfig, num_iters: int = 100) -> float:
    torch.cuda.empty_cache()
    dataloader = torch.utils.data.DataLoader(
        get_dataset(),
        batch_size=config.batch_size,
        num_workers=config.num_workers,
        pin_memory=config.pin_memory,
        persistent_workers=config.persistent_workers,
        prefetch_factor=config.prefetch_factor
    )
    has_started = False
    for i in range(num_iters):
        for images, labels in dataloader:
            if not has_started:  # only start timing after the first iteration
                has_started = True
                start_time = time.time()
            images.to('cuda')
            labels.to('cuda')
            torch.cuda.synchronize()
    end_time = time.time()
    time_per_epoch = (end_time - start_time) / num_iters
    print(f'Time taken: {time_per_epoch}')
    return time_per_epoch


class RepeatPruner(optuna.pruners.BasePruner):

    def prune(self, study: optuna.Study, trial: optuna.trial.FrozenTrial) -> bool:
        trials = study.get_trials(deepcopy=False)
        if any(other_trial.params == trial.params and other_trial.number < trial.number for other_trial in trials):
            return True
        return False


@app.command()
def run(
    config_paths: List[str] = typer.Option(..., help='Paths to config files'),
    sweep_paths: List[str] = typer.Option(default_factory=list, help='Paths to sweep files'),
    num_trials: int = typer.Option(50, help='Number of trials to run')
) -> None:
    dataloader_config = DataLoaderConfig.from_yaml_files(config_paths)

    if len(sweep_paths) == 0:
        dataloader_speed_benchmark(dataloader_config, num_iters=10)
        return

    sweep_config = DataLoaderConfigSweep.from_yaml_files(sweep_paths)

    def objective(trial: optuna.Trial) -> float:
        sampler = pgml.OptunaSampler(trial=trial)
        patch = sampler.sample(sweep_config)
        return dataloader_speed_benchmark(dataloader_config.merge(patch), num_iters=10)

    study = optuna.create_study(storage='sqlite:///sweep_dataloader.db', direction='minimize', pruner=RepeatPruner())
    study.optimize(objective, n_trials=num_trials)
    print(f'Best value: {study.best_value} (params: {study.best_params})')


if __name__ == '__main__':
    app()



import typing

import py_gen_ml as pgml

from . import mlp_patch as patch
from . import mlp_base as base

ActivationSweepField = typing.Union[
    pgml.Choice[base.Activation],
    typing.Literal['any'],
    base.Activation,
]


class MLPParsingDemoSweep(pgml.Sweeper[patch.MLPParsingDemoPatch]):
    """MLP is a simple multi-layer perceptron."""

    num_layers: typing.Optional[pgml.IntSweep] = None
    """Number of layers in the MLP."""

    num_units: typing.Optional[pgml.IntSweep] = None
    """Number of units in each layer."""

    activation: typing.Optional[ActivationSweepField] = None
    """Activation function to use."""



MLPParsingDemoSweepField = typing.Union[
    MLPParsingDemoSweep,
    pgml.NestedChoice[MLPParsingDemoSweep, patch.MLPParsingDemoPatch],  # type: ignore
]




# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml
import typing

import pydantic
import typer

from . import advanced_base as base


class TrainingArgs(pgml.YamlBaseModel):
    """Training configuration"""

    num_units: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of units. Maps to 'mlp.layers.num_units'"),
        pydantic.Field(None),
        pgml.ArgRef("mlp.layers.num_units"),
    ]
    """Number of units"""

    activation: typing.Annotated[
        typing.Optional[str],
        typer.Option(help="Activation function. Maps to 'mlp.layers.activation'"),
        pydantic.Field(None),
        pgml.ArgRef("mlp.layers.activation"),
    ]
    """Activation function"""

    type: typing.Annotated[
        typing.Optional[str],
        typer.Option(help="Type of optimizer. Maps to 'optimizer.type'"),
        pydantic.Field(None),
        pgml.ArgRef("optimizer.type"),
    ]
    """Type of optimizer"""

    learning_rate: typing.Annotated[
        typing.Optional[float],
        typer.Option(help="Learning rate. Maps to 'optimizer.learning_rate'"),
        pydantic.Field(None),
        pgml.ArgRef("optimizer.learning_rate"),
    ]
    """Learning rate"""



import logging


def setup_logger(name: str) -> logging.Logger:
    """
    Setup a logger with the given name.

    This function sets up a logger with the specified name and returns it.
    The logger is configured to log messages to the console at the INFO level.

    Args:
        name (str): The name of the logger.

    Returns:
        logging.Logger: The logger.
    """
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger



import sys
from pathlib import Path

import pytest
from pydantic import BaseModel

sys.path.append(str(Path(__file__).parent.absolute()))

from unittest.mock import Mock, patch

from typer.testing import CliRunner

from .pgml_out_test import (
    explicit_cli_arg_test_entrypoint,
    implicit_cli_arg_test_entrypoint,
    unit_base,
)
from .pgml_out_test.explicit_cli_arg_test_entrypoint import \
    app as explicit_cli_arg_test_app
from .pgml_out_test.implicit_cli_arg_test_entrypoint import \
    app as implicit_cli_arg_test_app

cli_runner = CliRunner()


def test_oneof() -> None:
    assert unit_base.OneofTest(value=1).value == 1
    assert unit_base.OneofTest(value='test').value == 'test'


def test_repeated() -> None:
    assert unit_base.RepeatedTest(values=[1, 2, 3]).values == [1, 2, 3]


def test_optional() -> None:
    assert unit_base.OptionalTest(value=1).value == 1
    assert unit_base.OptionalTest().value is None


def test_default() -> None:
    assert unit_base.Int32DefaultTest().value == 1
    assert unit_base.Int64DefaultTest().value == 1
    assert unit_base.Uint32DefaultTest().value == 1
    assert unit_base.Uint64DefaultTest().value == 1
    assert unit_base.Sint32DefaultTest().value == 1
    assert unit_base.Sint64DefaultTest().value == 1
    assert unit_base.Fixed32DefaultTest().value == 1
    assert unit_base.Fixed64DefaultTest().value == 1
    assert unit_base.Sfixed32DefaultTest().value == 1
    assert unit_base.Sfixed64DefaultTest().value == 1
    assert unit_base.FloatDefaultTest().value == 1
    assert unit_base.DoubleDefaultTest().value == 1
    assert unit_base.BoolDefaultTest().value == True
    assert unit_base.StringDefaultTest().value == 'test'
    assert unit_base.BytesDefaultTest().value == b'test'
    assert unit_base.EnumDefaultTest().value == unit_base.Enum.VALUE_1


def assert_base_model_eq(a: BaseModel, b: BaseModel) -> None:
    assert a.model_dump() == b.model_dump()


@pytest.fixture
def mocked_explicit_cli_arg_run_trial() -> Mock:
    with patch.object(explicit_cli_arg_test_entrypoint, 'run_trial') as mock_run_trial:
        yield mock_run_trial


def test_cli_arg_explicit_no_override(tmp_path: Path, mocked_explicit_cli_arg_run_trial: Mock) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text('bar: test')
    result = cli_runner.invoke(explicit_cli_arg_test_app, ['--config-paths', str(config_path)])
    assert result.exit_code == 0
    mocked_explicit_cli_arg_run_trial.assert_called_once()
    assert_base_model_eq(mocked_explicit_cli_arg_run_trial.call_args[0][0], unit_base.ExplicitCLIArgTest(bar='test'))


def test_cli_arg_explicit_override(tmp_path: Path, mocked_explicit_cli_arg_run_trial: Mock) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text('bar: test')
    result = cli_runner.invoke(explicit_cli_arg_test_app, ['--config-paths', str(config_path), '--foo', 'test2'])
    assert result.exit_code == 0
    mocked_explicit_cli_arg_run_trial.assert_called_once()
    assert_base_model_eq(mocked_explicit_cli_arg_run_trial.call_args[0][0], unit_base.ExplicitCLIArgTest(bar='test2'))


@pytest.fixture
def mocked_implicit_cli_arg_run_trial() -> Mock:
    with patch.object(implicit_cli_arg_test_entrypoint, 'run_trial') as mock_run_trial:
        yield mock_run_trial


def test_cli_arg_implicit_no_override(tmp_path: Path, mocked_implicit_cli_arg_run_trial: Mock) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text('bar: test')
    result = cli_runner.invoke(implicit_cli_arg_test_app, ['--config-paths', str(config_path)])
    assert result.exit_code == 0
    mocked_implicit_cli_arg_run_trial.assert_called_once()
    assert_base_model_eq(mocked_implicit_cli_arg_run_trial.call_args[0][0], unit_base.ImplicitCLIArgTest(bar='test'))


def test_cli_arg_implicit_override(tmp_path: Path, mocked_implicit_cli_arg_run_trial: Mock) -> None:
    config_path = tmp_path.joinpath('config.yaml')
    config_path.write_text('bar: test')
    result = cli_runner.invoke(implicit_cli_arg_test_app, ['--config-paths', str(config_path), '--bar', 'test2'])
    assert result.exit_code == 0
    mocked_implicit_cli_arg_run_trial.assert_called_once()
    assert_base_model_eq(mocked_implicit_cli_arg_run_trial.call_args[0][0], unit_base.ImplicitCLIArgTest(bar='test2'))



# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml


class DataLoaderConfig(pgml.YamlBaseModel):
    """DataLoader configuration"""

    batch_size: int
    """Batch size"""

    num_workers: int
    """Number of workers"""

    pin_memory: bool
    """Pin memory"""

    persistent_workers: bool
    """Persistent workers"""

    prefetch_factor: int
    """Prefetch factor"""



import enum
import inspect
from typing import Any, Dict, Literal, get_args

import optuna
from pydantic import BaseModel

from py_gen_ml.logging.setup_logger import setup_logger

logger = setup_logger(__name__)

from py_gen_ml.sweep.sweep import (
    Choice,
    FloatLogUniform,
    FloatUniform,
    IntUniform,
    NestedChoice,
    Sweeper,
    SweepModel,
    SweepSampler,
    SweepSamplerContext,
    SweepVisitor,
    TBaseModel,
    TScalar,
    TSweep,
)


class OptunaSampler(SweepSampler[TBaseModel]):
    """
    A sampler that uses Optuna for hyperparameter optimization.

    Args:
        trial (optuna.Trial): The optuna trial to use for hyperparameter optimization.
    """

    def __init__(self, trial: optuna.Trial) -> None:
        """
        Initialize the OptunaSampler.

        Args:
            trial (optuna.Trial): The optuna trial to use for hyperparameter optimization.
        """
        self._trial = trial
        self._context = SweepSamplerContext(path_parts=[], ctx=trial)
        self._visitor = OptunaVisitor()

    def sample(self, sweep: Sweeper[TBaseModel]) -> TBaseModel:
        """
        Sample a model from the sweep.

        Args:
            sweep (TSweep): The sweep to sample from.

        Returns:
            TBaseModel: The sampled model.
        """
        return sweep.accept(self._visitor, context=self._context)


class OptunaVisitor(SweepVisitor[optuna.Trial]):
    """
    A visitor that uses Optuna for hyperparameter optimization.

    Args:
        trial (optuna.Trial): The optuna trial to use for hyperparameter optimization.
    """

    def visit_float_log_uniform(
        self,
        log_uniform: FloatLogUniform,
        context: SweepSamplerContext[optuna.Trial],
    ) -> float:
        """
        Visit a float log uniform and resolve its value.

        Args:
            log_uniform (FloatLogUniform): The float log uniform to visit.
            context (SweepSamplerContext): The context for the sweep sampler.

        Returns:
            float: The resolved value.
        """
        return context.ctx.suggest_float(
            name=context.path,
            low=log_uniform.log_low,
            high=log_uniform.log_high,
            log=True,
        )

    def visit_int_uniform(self, uniform: IntUniform, context: SweepSamplerContext[optuna.Trial]) -> int:
        """
        Visit an int uniform and resolve its value.

        Args:
            uniform (IntUniform): The int uniform to visit.
            context (SweepSamplerContext): The context for the sweep sampler.

        Returns:
            int: The resolved value.
        """
        return context.ctx.suggest_int(
            name=context.path,
            low=uniform.low,
            high=uniform.high,
            step=uniform.step,
        )

    def visit_float_uniform(self, uniform: FloatUniform, context: SweepSamplerContext[optuna.Trial]) -> float:
        """
        Visit a float uniform and resolve its value.

        Args:
            uniform (FloatUniform): The float uniform to visit.
            context (SweepSamplerContext): The context for the sweep sampler.

        Returns:
            float: The resolved value.
        """
        return context.ctx.suggest_float(
            name=context.path,
            low=uniform.low,
            high=uniform.high,
            step=uniform.step,
        )

    def visit_nested_choice(
        self,
        sweep_choice: NestedChoice[TSweep, TBaseModel],
        context: SweepSamplerContext[optuna.Trial],
    ) -> TBaseModel:
        """
        Visit a nested choice and resolve its value.

        Args:
            sweep_choice (NestedChoice[TSweep, TBaseModel]): The nested choice to visit.
            context (SweepSamplerContext): The context for the sweep sampler.

        Returns:
            TBaseModel: The resolved base model.
        """
        random_choice = context.ctx.suggest_categorical(
            name=context.path,
            choices=list(sweep_choice.nested_options.keys()),
        )
        value = sweep_choice.nested_options[random_choice]
        return value.accept(self, context=context)  # type: ignore

    def visit_choice(self, choice: Choice[TScalar], context: SweepSamplerContext[optuna.Trial]) -> TScalar:
        """
        Visit a choice and resolve its value.

        Args:
            choice (Choice[TScalar]): The choice to visit.
            context (SweepSamplerContext): The context for the sweep sampler.

        Returns:
            TScalar: The resolved value.
        """
        return context.ctx.suggest_categorical(  # type: ignore
            name=context.path,
            choices=choice.options,  # type: ignore
        )

    def visit_sweep_model(
        self,
        sweep_model: Sweeper[TBaseModel],
        context: SweepSamplerContext[optuna.Trial],
    ) -> BaseModel:
        """
        Visit a sweep model and resolve its fields.

        Args:
            sweep_model (Sweeper[TBaseModel]): The sweep model to visit.
            context (SweepSamplerContext): The context for the sweep sampler.

        Returns:
            BaseModel: The resolved base model.
        """
        kwargs: Dict[str, Any] = {}
        for field_name, field in sweep_model.model_fields.items():
            field_value = getattr(sweep_model, field_name)
            if isinstance(field_value, SweepModel):
                kwargs[field_name] = field_value.accept(self, context=context.step(field_name))
            elif isinstance(field_value, list):
                resolved_list = []
                for index, item in enumerate(field_value):
                    if isinstance(item, SweepModel):
                        resolved_list.append(item.accept(self, context=context.step(field_name, index=index)))
                    else:
                        raise ValueError(f'SweepModel field {field_name} is not a Node')
            elif any(type_arg is Literal['any'] for type_arg in get_args(field.annotation)) and any(
                type_arg is bool or (inspect.isclass(type_arg) and issubclass(type_arg, enum.Enum))
                for type_arg in get_args(field.annotation)
            ):
                if field_value == 'any':
                    if any(type_arg is bool for type_arg in get_args(field.annotation)):
                        kwargs[field_name] = context.ctx.suggest_categorical(
                            name=context.step(field_name).path,
                            choices=[True, False],
                        )
                    elif any(
                        inspect.isclass(enum_class := type_arg) and issubclass(enum_class, enum.Enum)
                        for type_arg in get_args(field.annotation)
                    ):
                        kwargs[field_name] = context.ctx.suggest_categorical(
                            name=context.step(field_name).path,
                            choices=[enum_member.value for enum_member in enum_class],  # type: ignore
                        )
                    else:
                        raise ValueError(f'Unsupported type: {field.annotation}')
                else:
                    kwargs[field_name] = field_value
            else:
                kwargs[field_name] = field_value
        return sweep_model.new_base_model(**kwargs)



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class ModelPatch(pgml.YamlBaseModel):
    """Model configuration"""

    num_layers: typing.Optional[int] = None
    """Number of layers"""


class TrainingPatch(pgml.YamlBaseModel):
    """Training configuration"""

    num_epochs: typing.Optional[int] = None
    """Number of epochs"""


class DatasetPatch(pgml.YamlBaseModel):
    """Dataset configuration"""

    path: typing.Optional[str] = None
    """Path to the dataset"""


class DataPatch(pgml.YamlBaseModel):
    """Data config"""

    train_dataset: typing.Optional[DatasetPatch] = None
    """Path to the dataset"""

    test_dataset: typing.Optional[DatasetPatch] = None
    """Path to the dataset"""

    num_workers: typing.Optional[int] = None
    """Number of workers for loading the dataset"""


class CliExtensionDemoPatch(pgml.YamlBaseModel):
    """Global configuration"""

    data: typing.Optional[DataPatch] = None
    """Dataset configuration"""

    model: typing.Optional[ModelPatch] = None
    """Model configuration"""

    training: typing.Optional[TrainingPatch] = None
    """Training configuration"""



from typing import Final

BASE_MODEL_ALIAS: Final[str] = 'base'
SWEEP_MODEL_ALIAS: Final[str] = 'sweep'
PATCH_MODEL_ALIAS: Final[str] = 'patch'
CLI_ARGS_ALIAS: Final[str] = 'cli_args'
PGML_ALIAS: Final[str] = 'pgml'



from pathlib import Path
from typing import List, Set, TypeVar

import networkx
import protogen

from py_gen_ml.logging.setup_logger import setup_logger
from py_gen_ml.plugin.common import (
    field_to_default,
    generate_docstring,
    get_element_subgraphs,
    get_extension_value,
    py_import_for_source_file_derived_file,
    snake_case,
)
from py_gen_ml.plugin.constants import BASE_MODEL_ALIAS, PGML_ALIAS
from py_gen_ml.plugin.generator import Generator
from py_gen_ml.typing.some import some

logger = setup_logger(__name__)

T = TypeVar('T')


class BaseModelGenerator(Generator):
    """
    A generator that creates base models for a set of protobuf files.

    The generated file will be imported by the py-gen-ml CLI to generate base models for the given protobuf files.
    """

    def __init__(self, gen: protogen.Plugin, is_patch: bool, suffix: str) -> None:
        """
        Initialize the BaseModelGenerator.

        Args:
            gen (protogen.Plugin): The plugin instance.
            is_patch (bool): Whether this is a patch model.
            suffix (str): The suffix for the generated file.
        """
        super().__init__(gen)
        self._suffix = suffix
        self._is_patch = is_patch

    def _generate_code_for_file(self, file: protogen.File) -> None:
        """
        Generate base models for a specific file.

        This method creates a new generated file for the base models of the given file.
        It imports necessary modules and defines the base models for all messages and enums.
        """
        g = self._gen.new_generated_file(
            file.proto.name.replace('.proto', self._suffix),
            file.py_import_path,
        )
        g.P('# Autogenerated code. DO NOT EDIT.')
        if self._is_patch or self._requires_typing_import(file):
            g.P('import typing')
        if len(file.enums) > 0 and not self._is_patch:
            g.P('import enum')
            g.P()
        g.P(f'import py_gen_ml as {PGML_ALIAS}')
        g.P()
        if not self._is_patch and len(builder_imports := self.gather_factory_imports(file)) > 0:
            g.P('from typing import TYPE_CHECKING')
            g.P()
            g.P('if TYPE_CHECKING:')
            g.set_indent(4)
            for builder_import in builder_imports:
                g.P(f'import {builder_import}')
            g.set_indent(0)
            g.P()

        if self._is_patch and len(file.enums) > 0:
            import_file_path = str(Path(file.proto.name.replace('.proto', '_base')).name)
            g.P(f'from . import {import_file_path} as {BASE_MODEL_ALIAS}')
            g.P()
        g.P()

        dependency_subgraphs = get_element_subgraphs(file, include_elements={protogen.Kind.ENUM, protogen.Kind.MESSAGE})
        if not self._is_patch:
            for subgraph in dependency_subgraphs:
                for message in networkx.topological_sort(subgraph):
                    if not isinstance(message, protogen.Enum):
                        continue
                    self._generate_enum_base_model(g, message)

        for subgraph in dependency_subgraphs:
            for message in networkx.topological_sort(subgraph):
                if not isinstance(message, protogen.Message):
                    continue
                self._generate_message_base_model(g, message)

        self._run_yapf(g)

    def _generate_enum_base_model(self, g: protogen.GeneratedFile, enum: protogen.Enum) -> None:
        """
        Generate the base model for an enum.

        This method generates the class definition for the base model of the given enum.
        It includes the class name, inheritance, and docstring.
        """
        g.P(f'class {enum.proto.name}(str, enum.Enum):')
        g.set_indent(4)
        generate_docstring(g, enum)

        for value in enum.values:
            g.P(f"{value.proto.name} = \"{value.proto.name}\"")
            generate_docstring(g, value)

        g.set_indent(0)
        g.P()
        g.P()

    def _generate_message_base_model(self, g: protogen.GeneratedFile, message: protogen.Message) -> None:
        """
        Generate the base model for a message.

        This method generates the class definition for the base model of the given message.
        It includes the class name, inheritance, and docstring.
        """
        patch = 'Patch' if self._is_patch else ''
        g.P(f'class {message.proto.name}{patch}({PGML_ALIAS}.YamlBaseModel):')
        g.set_indent(4)
        generate_docstring(g, message)

        for field in message.fields:
            if field.oneof and len(field.oneof.fields) > 1:
                continue
            g.P(f'{field.py_name}: {self.field_to_annotation(field)}')
            generate_docstring(g, field)

        # Add oneof fields
        for oneof in message.oneofs:
            if len(oneof.fields) == 1:
                continue
            self._generate_oneof_field(g, oneof)

        if self._is_patch:
            g.set_indent(0)
            g.P()
            g.P()

            self._add_json_schema_gen_task(
                obj_path=py_import_for_source_file_derived_file(g, '_patch'),
                obj_name=message.py_ident.py_name + 'Patch',
                path=f'{self._configs_dir}/patch/schemas/' + snake_case(message.proto.name) + '.json',
            )
            return

        if (builder_import_path := get_extension_value(message, 'factory', str)) is not None:
            dot_index = builder_import_path.rfind('.')
            if dot_index == -1:
                raise ValueError(f'Invalid builder import path: {builder_import_path}')

            g.P()
            g.P("def build(self) -> \"", builder_import_path, "\":")

            g.set_indent(8)
            if (idx := builder_import_path.rfind('.')) == -1:
                raise ValueError(f'Invalid builder import {builder_import_path}')
            g.P('import ', builder_import_path[:idx])
            g.P()
            g.P('return ', builder_import_path, '(')

            g.set_indent(12)
            for field in message.fields:
                if ((field_message := field.message) is not None and
                    get_extension_value(field_message, 'factory', str) is not None):
                    build_field = True
                else:
                    build_field = False

                as_varargs = get_extension_value(field, 'as_varargs', bool) or False
                is_list = field.is_list()
                if as_varargs and build_field and is_list:
                    g.P(f'*(elem.build() for elem in self.{field.py_name}),')
                elif not as_varargs and build_field and is_list:
                    g.P(f'{field.py_name}=[elem.build() for elem in self.{field.py_name}],')
                elif not as_varargs and build_field and not is_list:
                    g.P(f'{field.py_name}=self.{field.py_name}.build(),')
                elif not as_varargs and not build_field:
                    g.P(field.py_name, '=self.', field.py_name, ',')
                else:
                    raise ValueError(f'Unsupported case {message.full_name}.{field.py_name}')

            g.set_indent(8)
            g.P(')')

            g.set_indent(4)

        g.set_indent(0)
        g.P()
        g.P()

        self._add_json_schema_gen_task(
            obj_path=py_import_for_source_file_derived_file(g, '_base'),
            obj_name=message.py_ident.py_name,
            path=f'{self._configs_dir}/base/schemas/' + snake_case(message.proto.name) + '.json',
        )

    def _generate_oneof_field(self, g: protogen.GeneratedFile, oneof: protogen.OneOf) -> None:
        """
        Generate the base model for a oneof field.

        This method generates the class definition for the base model of the given oneof field.
        It includes the class name, inheritance, and docstring.

        Args:
            g (protogen.GeneratedFile): The generated file to write to.
            oneof (protogen.OneOf): The oneof field to generate the base model for.
        """
        oneof_name = oneof.proto.name
        g.P(f'{oneof_name}: {self.oneof_to_annotation(oneof)}')
        generate_docstring(g, oneof)

    def oneof_to_annotation(self, oneof: protogen.OneOf) -> str:
        """
        Convert the oneof field to its corresponding annotation.

        This method determines the appropriate annotation for the given oneof field.
        It returns the annotation as a string.

        Args:
            oneof (protogen.OneOf): The oneof field to convert to an annotation.

        Returns:
            str: The annotation for the oneof field.
        """
        if len(oneof.fields) == 1:
            return self.field_to_type(oneof.fields[0])
        types = [self.field_to_type(field) for field in oneof.fields]
        union_type = ', '.join(types)
        return f'typing.Union[{union_type}]'

    @staticmethod
    def gather_factory_imports(file: protogen.File) -> List[str]:
        """
        Gather the builder imports for a specific file.

        This method collects the import paths for the factories of all messages in the given file.
        It ensures that each builder import path is unique and returns a sorted list of these paths.

        Args:
            file (protogen.File): The file to gather builder imports from.

        Returns:
            List[str]: A sorted list of unique builder import paths.
        """
        imports: Set[str] = set()
        for message in file.messages:
            factory_import_path = get_extension_value(message, 'factory', str)
            if factory_import_path is None:
                continue

            if (idx := factory_import_path.rfind('.')) == -1:
                raise ValueError(f'Invalid factory import {factory_import_path}')
            imports.add(factory_import_path[:idx])
        return sorted(imports)

    def field_to_annotation(self, field: protogen.Field) -> str:
        """
        Convert the field to its corresponding annotation.

        This method determines the appropriate annotation for the given field based on its type.
        It returns the annotation as a string.

        Args:
            field (protogen.Field): The field to convert to an annotation.

        Returns:
            str: The annotation for the field.
        """
        annotation = self.field_to_type(field)
        default = field_to_default(field)
        if field.is_list():
            annotation = f'typing.List[{annotation}]'
        if field.proto.proto3_optional or self._is_patch:
            annotation = f'typing.Optional[{annotation}]'
            if default is None or self._is_patch:
                default = 'None'

        if default is not None:
            return f'{annotation} = {default}'

        return annotation

    def field_to_type(self, field: protogen.Field) -> str:
        """
        Convert the field to its corresponding type.

        This method determines the appropriate type for the given field based on its kind.
        It returns the type as a string.

        Args:
            field (protogen.Field): The field to convert to a type.

        Returns:
            str: The type for the field.
        """
        if field.kind == protogen.Kind.MESSAGE:
            message = some(field.message)
            if self._is_patch:
                return message.py_ident.py_name + 'Patch'
            return message.py_ident.py_name
        elif field.kind == protogen.Kind.DOUBLE:
            return 'float'
        elif field.kind == protogen.Kind.FLOAT:
            return 'float'
        elif field.kind == protogen.Kind.INT64:
            return 'int'
        elif field.kind == protogen.Kind.UINT64:
            return 'int'
        elif field.kind == protogen.Kind.INT32:
            return 'int'
        elif field.kind == protogen.Kind.FIXED64:
            return 'int'
        elif field.kind == protogen.Kind.FIXED32:
            return 'int'
        elif field.kind == protogen.Kind.BOOL:
            return 'bool'
        elif field.kind == protogen.Kind.STRING:
            return 'str'
        elif field.kind == protogen.Kind.BYTES:
            return 'bytes'
        elif field.kind == protogen.Kind.UINT32:
            return 'int'
        elif field.kind == protogen.Kind.ENUM:
            enum = some(field.enum)
            if self._is_patch:
                return f'{BASE_MODEL_ALIAS}.{enum.py_ident.py_name}'
            return enum.py_ident.py_name
        elif field.kind == protogen.Kind.SFIXED32:
            return 'int'
        elif field.kind == protogen.Kind.SFIXED64:
            return 'int'
        elif field.kind == protogen.Kind.SINT32:
            return 'int'
        elif field.kind == protogen.Kind.SINT64:
            return 'int'
        else:
            raise ValueError(f'Unknown field kind: {field.kind}')



# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml
import typing

import pydantic
import typer

from . import cli_demo_base as base


class CLIDemoArgs(pgml.YamlBaseModel):
    """Global configuration"""

    path: typing.Annotated[
        typing.Optional[str],
        typer.Option(help="Path to the dataset. Maps to 'data.dataset.path'"),
        pydantic.Field(None),
        pgml.ArgRef("data.dataset.path"),
    ]
    """Path to the dataset"""

    num_layers: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of layers. Maps to 'model.num_layers'"),
        pydantic.Field(None),
        pgml.ArgRef("model.num_layers"),
    ]
    """Number of layers"""

    num_workers: typing.Annotated[
        typing.Optional[int],
        typer.Option(
            help="Number of workers for loading the dataset. Maps to 'data.num_workers'"
        ),
        pydantic.Field(None),
        pgml.ArgRef("data.num_workers"),
    ]
    """Number of workers for loading the dataset"""

    num_epochs: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of epochs. Maps to 'training.num_epochs'"),
        pydantic.Field(None),
        pgml.ArgRef("training.num_epochs"),
    ]
    """Number of epochs"""



import typing

import py_gen_ml as pgml

from . import oneof_demo_patch as patch
from . import oneof_demo_base as base


class TransformerSweep(pgml.Sweeper[patch.TransformerPatch]):
    """Transformer configuration"""

    num_layers: typing.Optional[pgml.IntSweep] = None
    """Number of layers"""

    num_heads: typing.Optional[pgml.IntSweep] = None
    """Number of heads"""

    activation: typing.Optional[pgml.StrSweep] = None
    """Activation function"""


TransformerSweepField = typing.Union[
    TransformerSweep,
    pgml.NestedChoice[TransformerSweep, patch.TransformerPatch],  # type: ignore
]


class ConvBlockSweep(pgml.Sweeper[patch.ConvBlockPatch]):
    """Conv block"""

    out_channels: typing.Optional[pgml.IntSweep] = None
    """Number of output channels"""

    kernel_size: typing.Optional[pgml.IntSweep] = None
    """Kernel size"""

    activation: typing.Optional[pgml.StrSweep] = None
    """Activation function"""


ConvBlockSweepField = typing.Union[
    ConvBlockSweep,
    pgml.NestedChoice[ConvBlockSweep, patch.ConvBlockPatch],  # type: ignore
]


class ConvNetSweep(pgml.Sweeper[patch.ConvNetPatch]):
    """Convolutional neural network configuration"""

    layers: typing.Optional[ConvBlockSweepField] = None
    """Conv layer configuration"""


ConvNetSweepField = typing.Union[
    ConvNetSweep,
    pgml.NestedChoice[ConvNetSweep, patch.ConvNetPatch],  # type: ignore
]


class ModelSweep(pgml.Sweeper[patch.ModelPatch]):
    """Model configuration"""

    backbone: typing.Optional[typing.Union[pgml.Sweeper[TransformerSweepField],
                                           pgml.Sweeper[ConvNetSweepField]]] = None



ModelSweepField = typing.Union[
    ModelSweep,
    pgml.NestedChoice[ModelSweep, patch.ModelPatch],  # type: ignore
]




from typing import Optional, TypeVar

T = TypeVar('T')


def some(value: Optional[T]) -> T:
    """
    Get the value if it is not None.

    Args:
        value (Optional[T]): The value to get.

    Raises:
        ValueError: If the value is None.

    Returns:
        T: The value.
    """
    if value is None:
        raise ValueError('Value is None')
    return value



import re
import textwrap
import typing
from typing import Final, Literal, Optional, Set, TypeVar, Union

import google.protobuf.json_format
import more_itertools
import networkx
import protogen
from google.protobuf.descriptor import FieldDescriptor

from py_gen_ml.extensions_pb2 import FieldDefaults
from py_gen_ml.logging.setup_logger import setup_logger
from py_gen_ml.typing.some import some

logger = setup_logger(__name__)

WRAP_WIDTH: Final[int] = 84

T = TypeVar('T')


def get_extension_value(
    element: Union[protogen.Field, protogen.Message, protogen.File],
    extension_name: str,
    extension_type: type[T],
) -> Optional[T]:
    """
    Get the value of an extension field from the proto options.

    This function searches for an extension field with the given name in the
    options of the provided element (which can be a Field, Message, or File).
    If found, it returns the value of the extension field. Otherwise, it returns None.

    Args:
        element (typing.Union[protogen.Field, protogen.Message, protogen.File]): The element to search for the extension field.
        extension_name (str): The name of the extension field to search for.
        extension_type (type[T]): The type of the extension field.

    Returns:
        Optional[T]: The value of the extension field if found, otherwise None.
    """
    option: typing.Optional[tuple[FieldDescriptor, T]] = more_itertools.first_true(
        element.proto.options.ListFields(),
        None,
        lambda x: x[0].name == extension_name,
    )

    if option is None:
        return None
    return option[1]


def field_to_default(field: protogen.Field, import_prefix: str = '') -> typing.Optional[str]:
    """
    Convert the default value of a field to its corresponding Python representation.

    This function retrieves the default value of a field using the `FieldDefaults`
    extension and converts it to its Python string representation. The conversion
    depends on the type of the field (e.g., int, float, string, etc.).

    Args:
        field (protogen.Field): The field to get the default value for.
        import_prefix (str): The prefix to use for imports in the generated code.

    Returns:
        Optional[str]: The default value of the field as a Python string, or None if no default is set.
    """
    proto_default = get_extension_value(field, 'default', FieldDefaults)
    if proto_default is None:
        return None
    if field.kind == protogen.Kind.DOUBLE:
        default_field_name = 'double'
        default_value = proto_default.double
    elif field.kind == protogen.Kind.FLOAT:
        default_field_name = 'float'
        default_value = proto_default.float
    elif field.kind == protogen.Kind.INT64:
        default_field_name = 'int64'
        default_value = proto_default.int64
    elif field.kind == protogen.Kind.UINT64:
        default_field_name = 'uint64'
        default_value = proto_default.uint64
    elif field.kind == protogen.Kind.INT32:
        default_field_name = 'int32'
        default_value = proto_default.int32
    elif field.kind == protogen.Kind.FIXED64:
        default_field_name = 'fixed64'
        default_value = proto_default.fixed64
    elif field.kind == protogen.Kind.FIXED32:
        default_field_name = 'fixed32'
        default_value = proto_default.fixed32
    elif field.kind == protogen.Kind.BOOL:
        default_field_name = 'bool'
        default_value = proto_default.bool
    elif field.kind == protogen.Kind.STRING:
        default_field_name = 'string'
        default_value = f"\"{proto_default.string}\""
    elif field.kind == protogen.Kind.BYTES:
        default_field_name = 'bytes'
        default_value = proto_default.bytes
    elif field.kind == protogen.Kind.UINT32:
        default_field_name = 'uint32'
        default_value = proto_default.uint32
    elif field.kind == protogen.Kind.SFIXED32:
        default_field_name = 'sfixed32'
        default_value = proto_default.sfixed32
    elif field.kind == protogen.Kind.SFIXED64:
        default_field_name = 'sfixed64'
        default_value = proto_default.sfixed64
    elif field.kind == protogen.Kind.SINT32:
        default_field_name = 'sint32'
        default_value = proto_default.sint32
    elif field.kind == protogen.Kind.SINT64:
        default_field_name = 'sint64'
        default_value = proto_default.sint64
    elif field.kind == protogen.Kind.ENUM:
        default_field_name = 'enum'
        enum = some(field.enum)
        names = [v.name for v in enum.proto.value]
        if proto_default.enum not in names:
            raise ValueError(f'Invalid enum value: {proto_default.enum}')
        default_value = f'{import_prefix}{enum.py_ident.py_name}.{proto_default.enum}'
    else:
        raise ValueError(f'Default not supported for kind: {field.kind}')
    if field.kind in {
        protogen.Kind.FLOAT,
        protogen.Kind.DOUBLE,
    }:
        as_json = google.protobuf.json_format.MessageToJson(field.proto.options, indent=None)
        pattern = re.compile(rf'"{default_field_name}"\s*:\s*([+-]?\d*(?:\.\d+)?)')
        default_value = some(pattern.search(as_json)).group(1)
    return str(default_value) if default_value is not None else None


def generate_docstring(
    g: protogen.GeneratedFile,
    element: Union[protogen.Field, protogen.Message, protogen.Enum, protogen.EnumValue, protogen.OneOf],
) -> None:
    """
    Generate a docstring for the given element.

    This function generates a docstring for the provided element (which can be a Field, Message, Enum, or EnumValue).
    It wraps the leading comments of the element and formats them as a Python docstring.

    Args:
        g (protogen.GeneratedFile): The generated file to write the docstring to.
        element (Union[protogen.Field, protogen.Message, protogen.Enum, protogen.EnumValue]): The element to generate a docstring for.
    """
    if not element.location.leading_comments:
        return

    if WRAP_WIDTH - 6 - 4 >= len(element.location.leading_comments):
        g.P(f'"""{element.location.leading_comments}"""')
    else:
        g.P('"""')
        lines = textwrap.wrap(element.location.leading_comments, width=WRAP_WIDTH - 4)
        for line in lines:
            g.P(line)
        g.P('"""')
    g.P()


def get_element_subgraphs(
    file: protogen.File,
    include_elements: Optional[Set[Literal[protogen.Kind.MESSAGE, protogen.Kind.ENUM]]] = None,
) -> list[networkx.MultiDiGraph]:
    """
    Get the message subgraphs from a file.

    This function creates a multi-directed graph of messages and their dependencies.
    It adds nodes for each message and edges for each field that is a message type.
    The function then returns the connected components of the graph as subgraphs.

    Args:
        file (protogen.File): The file to get the message subgraphs from.

    Returns:
        List[networkx.MultiDiGraph]: The list of message subgraphs.
    """
    if include_elements is None:
        include_elements = {protogen.Kind.MESSAGE}
    graph = networkx.MultiDiGraph()
    for message in file.messages:
        graph.add_node(message)
        for field in message.fields:
            if field.kind not in include_elements:
                continue
            field_name = field.py_name
            if field.oneof is not None:
                field_name = field.oneof.proto.name
            if field.kind == protogen.Kind.MESSAGE:
                graph.add_edge(field.message, message, key=field_name)
            elif field.kind == protogen.Kind.ENUM:
                graph.add_edge(field.enum, message, key=field_name)
            else:
                raise ValueError(f'Unsupported field kind: {field.kind}')
    return [graph.subgraph(c).copy() for c in networkx.connected_components(graph.to_undirected())]  # type: ignore


def snake_case(name: str) -> str:
    """
    Convert a string to snake case.

    This function converts the given string to snake case by:
    1. Inserting underscores between transitions from lowercase to uppercase
    2. Inserting underscores between transitions from uppercase to lowercase (if not at the start of a word)
    3. Replacing spaces and punctuation with underscores
    4. Converting all characters to lowercase

    Args:
        name (str): The string to convert to snake case.

    Returns:
        str: The snake case version of the input string.
    """
    # Insert underscore between lowercase and uppercase transitions
    name = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', name)

    # Insert underscore between uppercase and lowercase transitions (not at the start of a word)
    name = re.sub(r'([A-Z])([A-Z][a-z])', r'\1_\2', name)

    # Replace any remaining non-alphanumeric characters with underscores
    name = re.sub(r'[^a-zA-Z0-9]+', '_', name)

    # Convert to lowercase
    return name.lower()


def py_import_for_source_file_derived_file(file: protogen.GeneratedFile, name: str) -> str:
    """
    Get the Python import path for a file derived from a source file.

    This function returns the Python import path for a file that is derived from the
    given source file. The derived file is typically a version of the source file
    with a different extension or suffix.

    Args:
        file (protogen.File): The source file.
        name (str): The name of the derived file.

    Returns:
        str: The Python import path for the derived file.
    """
    return file._py_import_path._path.replace('_pb2', f'{name}')


def field_requires_typing_import(field: protogen.Field) -> bool:
    """
    Determine if field requires a typing import.

    Will be True if any of the following is true:
    - Field is part of a oneof
    - Field is optional
    - Field is repeated

    Args:
        field (protogen.Field): Field to check.

    Returns:
        bool: True if annotating the field requires a typing import
    """
    if field.oneof is not None:
        return True
    if field.proto.proto3_optional:
        return True
    if field.is_list():
        return True
    return False



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class OptimizerPatch(pgml.YamlBaseModel):
    """Optimizer configuration"""

    type: typing.Optional[str] = None
    """Optimizer type"""

    learning_rate: typing.Optional[float] = None
    """Learning rate"""



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class DataLoaderConfigPatch(pgml.YamlBaseModel):
    """DataLoader configuration"""

    batch_size: typing.Optional[int] = None
    """Batch size"""

    num_workers: typing.Optional[int] = None
    """Number of workers"""

    pin_memory: typing.Optional[bool] = None
    """Pin memory"""

    persistent_workers: typing.Optional[bool] = None
    """Persistent workers"""

    prefetch_factor: typing.Optional[int] = None
    """Prefetch factor"""



import typing

import py_gen_ml as pgml

from . import builder_custom_class_demo_patch as patch
from . import builder_custom_class_demo_base as base


class LinearBlockSweep(pgml.Sweeper[patch.LinearBlockPatch]):
    """Linear block configuration"""

    in_features: typing.Optional[pgml.IntSweep] = None
    """Number of input features"""

    out_features: typing.Optional[pgml.IntSweep] = None
    """Number of output features"""

    bias: typing.Optional[pgml.BoolSweep] = None
    """Bias"""

    dropout: typing.Optional[pgml.FloatSweep] = None
    """Dropout probability"""

    activation: typing.Optional[pgml.StrSweep] = None
    """Activation function"""


LinearBlockSweepField = typing.Union[
    LinearBlockSweep,
    pgml.NestedChoice[LinearBlockSweep, patch.LinearBlockPatch],  # type: ignore
]


class MLPSweep(pgml.Sweeper[patch.MLPPatch]):
    """MLP configuration"""

    layers: typing.Optional[LinearBlockSweepField] = None
    """Linear blocks"""



MLPSweepField = typing.Union[
    MLPSweep,
    pgml.NestedChoice[MLPSweep, patch.MLPPatch],  # type: ignore
]




# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml

from . import config_base as base


class OptimizerPatch(pgml.YamlBaseModel):
    """Optimizer configuration"""

    learning_rate: float | None = None
    """Learning rate"""

    beta1: float | None = None
    """Decay rate"""


class DataPatch(pgml.YamlBaseModel):
    """Data configuration"""

    batch_size: int | None = None
    """Batch size for a single GPU"""

    num_epochs: int | None = None
    """Number of epochs to train"""


class ConvBlockPatch(pgml.YamlBaseModel):
    """Convolutional layer configuration"""

    out_channels: int | None = None
    """Number of output channels"""

    kernel_size: int | None = None
    """Square kernel size"""

    pool_size: int | None = None
    """Square pool size"""

    activation: base.Activation | None = None
    """Activation function"""


class LinearBlockPatch(pgml.YamlBaseModel):
    """Linear layer configuration"""

    out_features: int | None = None
    """Number of output features"""

    activation: base.Activation | None = None
    """Activation function"""


class ConvNetPatch(pgml.YamlBaseModel):
    """Convolutional neural network configuration"""

    block: ConvBlockPatch | None = None
    """Conv layer configuration"""

    num_layers: int | None = None
    """Number of layers"""


class MLPPatch(pgml.YamlBaseModel):
    """Multi-layer perceptron configuration"""

    block: LinearBlockPatch | None = None
    """Linear layer configuration"""

    num_layers: int | None = None
    """Number of layers"""


class ModelPatch(pgml.YamlBaseModel):
    """Model configuration"""

    conv_net: ConvNetPatch | None = None
    """Conv blocks"""

    head: MLPPatch | None = None
    """MLP head"""


class ProjectPatch(pgml.YamlBaseModel):
    """Global configuration"""

    net: ModelPatch | None = None
    """Model configuration"""

    optimizer: OptimizerPatch | None = None
    """Optimizer configuration"""

    data: DataPatch | None = None
    """Data configuration"""



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class LinearBlock(pgml.YamlBaseModel):
    """Linear block configuration"""

    num_units: int
    """Number of units"""

    activation: str
    """Activation function"""


class Optimizer(pgml.YamlBaseModel):
    """Optimizer configuration"""

    type: str
    """Type of optimizer"""

    learning_rate: float
    """Learning rate"""


class MLP(pgml.YamlBaseModel):
    """Multi-layer perceptron configuration"""

    layers: typing.List[LinearBlock]
    """List of linear blocks"""


class Training(pgml.YamlBaseModel):
    """Training configuration"""

    mlp: MLP
    """Multi-layer perceptron configuration"""

    optimizer: Optimizer
    """Optimizer configuration"""



from pathlib import Path
from typing import TypeVar

import networkx
import protogen

from py_gen_ml.plugin.common import (
    generate_docstring,
    get_element_subgraphs,
    py_import_for_source_file_derived_file,
    snake_case,
)
from py_gen_ml.plugin.constants import (
    BASE_MODEL_ALIAS,
    PATCH_MODEL_ALIAS,
    PGML_ALIAS,
)
from py_gen_ml.plugin.generator import Generator
from py_gen_ml.typing.some import some

T = TypeVar('T')


class SweepModelGenerator(Generator):
    """
    Initialize the SweepModelGenerator.

    Args:
        gen (protogen.Plugin): The protogen plugin instance.
        suffix (str): The suffix to be added to the generated file names.

    This class is responsible for generating sweep models based on the input protobuf files.
    Sweep models are used to parameterize the values of fields in a message.
    """

    def __init__(self, gen: protogen.Plugin, suffix: str) -> None:
        """
        Initialize the SweepModelGenerator.

        Args:
            gen (protogen.Plugin): The protogen plugin instance.
            suffix (str): The suffix to be added to the generated file names.

        This class is responsible for generating sweep models based on the input protobuf files.
        Sweep models are used to parameterize the values of fields in a message.
        """
        super().__init__(gen)
        self._suffix = suffix

    def _generate_code_for_file(self, file: protogen.File) -> None:
        """
        Generate sweep models for a specific file.

        This method creates a new generated file for the sweep models of the given file.
        It imports necessary modules and defines the sweep models for all messages and enums.
        """
        g = self._gen.new_generated_file(
            file.proto.name.replace('.proto', self._suffix),
            file.py_import_path,
        )
        g.P('import typing')
        g.P()
        g.P(f'import py_gen_ml as {PGML_ALIAS}')
        g.P()

        import_file_path = str(Path(file.proto.name.replace('.proto', '_patch')).name)
        g.P(f'from . import {import_file_path} as {PATCH_MODEL_ALIAS}')

        import_file_path = str(Path(file.proto.name.replace('.proto', '_base')).name)
        g.P(f'from . import {import_file_path} as {BASE_MODEL_ALIAS}')

        g.P()
        g.P()

        dependency_subgraphs = get_element_subgraphs(file, include_elements={protogen.Kind.ENUM, protogen.Kind.MESSAGE})

        for subgraph in dependency_subgraphs:
            for message in networkx.topological_sort(subgraph):
                if not isinstance(message, protogen.Enum):
                    continue
                self._generate_sweep_field_for_enum(g, message)

        for subgraph in dependency_subgraphs:
            for message in networkx.topological_sort(subgraph):
                if not isinstance(message, protogen.Message):
                    continue
                self._generate_sweep_model_for_message(g, message)
                g.P()
                g.P()
                self._generate_sweep_field_for_message(g, message)

    def _generate_sweep_model_for_message(self, g: protogen.GeneratedFile, message: protogen.Message) -> None:
        """
        Generate the sweep model for a specific message.

        This method generates the class definition for the sweep model of the given message.
        It includes the class name, inheritance, and docstring.

        Args:
            g (protogen.GeneratedFile): The generated file to write the sweep model to.
            message (protogen.Message): The message to generate the sweep model for.
        """
        prefix = PATCH_MODEL_ALIAS + '.'
        g.P(f'class {message.proto.name}Sweep({PGML_ALIAS}.Sweeper[{prefix}{message.proto.name}Patch]):')
        g.set_indent(4)
        generate_docstring(g, message)

        for field in message.fields:
            if field.oneof:
                continue
            g.P(f'{field.py_name}: {self.field_to_sweep_annotation(field)}')
            generate_docstring(g, field)

        # Add oneof fields
        for oneof in message.oneofs:
            self._generate_oneof_field(g, oneof)

        g.set_indent(0)

        self._add_json_schema_gen_task(
            obj_path=py_import_for_source_file_derived_file(g, '_sweep'),
            obj_name=f'{message.proto.name}Sweep',
            path=f'{self._configs_dir}/sweep/schemas/' + snake_case(message.proto.name) + '.json',
        )

        self._run_yapf(g)

    def _generate_sweep_field_for_enum(self, g: protogen.GeneratedFile, enum: protogen.Enum) -> None:
        """
        Generate the sweep field for a specific enum.

        This method generates the union type for the sweep field of the given enum.
        It includes the union of the enum sweep, fixed, and base model.

        Args:
            g (protogen.GeneratedFile): The generated file to write the sweep field to.
            enum (protogen.Enum): The enum to generate the sweep field for.
        """
        base_model = f'{BASE_MODEL_ALIAS}.{enum.proto.name}'

        g.P(f'{enum.proto.name}SweepField = typing.Union[')
        g.set_indent(4)
        g.P(f'{PGML_ALIAS}.Choice[{base_model}],')
        g.P(f"typing.Literal['any'],")
        g.P(f'{base_model},')
        g.set_indent(0)
        g.P(']')
        g.P()
        g.P()

    def _generate_sweep_field_for_message(self, g: protogen.GeneratedFile, message: protogen.Message) -> None:
        """
        Generate the sweep field for a specific message.

        This method generates the union type for the sweep field of the given message.
        It includes the union of the message sweep, nested choice, named choice, fixed, and base model.

        Args:
            g (protogen.GeneratedFile): The generated file to write the sweep field to.
            message (protogen.Message): The message to generate the sweep field for.
        """
        base_model = f'{message.proto.name}'

        base_model = f'{PATCH_MODEL_ALIAS}.{base_model}Patch'
        g.P(f'{message.proto.name}SweepField = typing.Union[')
        g.set_indent(4)
        g.P(f'{message.proto.name}Sweep,')
        g.P(f'{PGML_ALIAS}.NestedChoice[{message.proto.name}Sweep, {base_model}],  # type: ignore')
        g.set_indent(0)
        g.P(']')
        g.P()
        g.P()

    def field_to_sweep_annotation(self, field: protogen.Field) -> str:
        """
        Convert the field to its corresponding sweep annotation.

        This method determines the appropriate sweep annotation for the given field based on its type.
        It returns the sweep annotation as a string.

        Args:
            field (protogen.Field): The field to convert to its sweep annotation.

        Returns:
            str: The sweep annotation as a string.
        """
        annotation = self.field_to_sweep_type(field)
        return f'typing.Optional[{annotation}] = None'

    def field_to_sweep_type(self, field: protogen.Field) -> str:
        """
        Convert the field to its corresponding sweep type.

        This method determines the appropriate sweep type for the given field based on its type.
        It returns the sweep type as a string.

        Args:
            field (protogen.Field): The field to convert to its sweep type.

        Returns:
            str: The sweep type as a string.
        """
        if field.kind == protogen.Kind.MESSAGE:
            return some(field.message).py_ident.py_name + 'SweepField'
        elif field.kind == protogen.Kind.ENUM:
            return some(field.enum).py_ident.py_name + 'SweepField'
        elif field.kind == protogen.Kind.DOUBLE:
            return f'{PGML_ALIAS}.FloatSweep'
        elif field.kind == protogen.Kind.FLOAT:
            return f'{PGML_ALIAS}.FloatSweep'
        elif field.kind == protogen.Kind.INT64:
            return f'{PGML_ALIAS}.IntSweep'
        elif field.kind == protogen.Kind.UINT64:
            return f'{PGML_ALIAS}.IntSweep'
        elif field.kind == protogen.Kind.INT32:
            return f'{PGML_ALIAS}.IntSweep'
        elif field.kind == protogen.Kind.FIXED64:
            return f'{PGML_ALIAS}.IntSweep'
        elif field.kind == protogen.Kind.FIXED32:
            return f'{PGML_ALIAS}.IntSweep'
        elif field.kind == protogen.Kind.BOOL:
            return f'{PGML_ALIAS}.BoolSweep'
        elif field.kind == protogen.Kind.STRING:
            return f'{PGML_ALIAS}.StrSweep'
        elif field.kind == protogen.Kind.BYTES:
            return f'{PGML_ALIAS}.BytesSweep'
        elif field.kind == protogen.Kind.UINT32:
            return f'{PGML_ALIAS}.IntSweep'
        elif field.kind == protogen.Kind.SFIXED32:
            return f'{PGML_ALIAS}.IntSweep'
        elif field.kind == protogen.Kind.SFIXED64:
            return f'{PGML_ALIAS}.IntSweep'
        elif field.kind == protogen.Kind.SINT32:
            return f'{PGML_ALIAS}.IntSweep'
        elif field.kind == protogen.Kind.SINT64:
            return f'{PGML_ALIAS}.IntSweep'
        else:
            raise ValueError(f'Unknown field kind: {field.kind}')

    def _generate_oneof_field(self, g: protogen.GeneratedFile, oneof: protogen.OneOf) -> None:
        oneof_name = oneof.proto.name
        g.P(f'{oneof_name}: {self._oneof_to_annotation(oneof)}')
        generate_docstring(g, oneof)

    def _oneof_to_annotation(self, oneof: protogen.OneOf) -> str:
        types = [f'{PGML_ALIAS}.Sweeper[{self.field_to_sweep_type(field)}]' for field in oneof.fields]
        union_type = ', '.join(types)
        return f'typing.Optional[typing.Union[{union_type}]] = None'



# Autogenerated code. DO NOT EDIT.
import typing
import py_gen_ml as pgml


class LinearBlockPatch(pgml.YamlBaseModel):
    """Linear block configuration"""

    in_features: typing.Optional[int] = None
    """Number of input features"""

    out_features: typing.Optional[int] = None
    """Number of output features"""

    bias: typing.Optional[bool] = None
    """Bias"""

    dropout: typing.Optional[float] = None
    """Dropout probability"""

    activation: typing.Optional[str] = None
    """Activation function"""


class MLPPatch(pgml.YamlBaseModel):
    """MLP configuration"""

    layers: typing.Optional[typing.List[LinearBlockPatch]] = None
    """Linear blocks"""



import pathlib
import warnings
from typing import Any, Dict, List

with warnings.catch_warnings():
    warnings.filterwarnings('ignore', category=DeprecationWarning, module='jsonmerge')
    import jsonmerge

from pydantic import BaseModel
from ruamel.yaml import YAML
from typing_extensions import Self

from py_gen_ml.cli_args.arg_refs import apply_args
from py_gen_ml.yaml.object_path import (
    EXTERNAL_OBJECT_PATH_REGEX,
    INTERNAL_OBJECT_PATH_REGEX,
    ObjectPath,
    resolve_inner_ref,
)

yaml = YAML()


class YamlBaseModel(BaseModel):
    """
    A base model that can be used to create instances of models from yaml files.

    This class augments the Pydantic base model with the ability to load data from yaml files. Moreover,
    it offers the ability to merge multiple yaml files into a single model. It also internally
    resolves references to other objects in the yaml files. Lastly, it offers the ability to
    apply cli arguments to the model.

    !!! info
        All Pydantic models that are generated using `py-gen-ml` will inherit from this class. So make sure to
        read the documentation on this class to fully leverage all the features of `py-gen-ml`.
    """

    @classmethod
    def from_yaml_file(cls, path: str) -> Self:
        """
        Create a new instance of the model from a yaml file.

        Args:
            path (str): The path to the yaml file.

        Returns:
            The new instance of the model.
        """
        return cls.from_yaml_files([path])

    @classmethod
    def load_data_with_references(cls, path: str) -> Dict[str, Any]:
        """
        Load data from a yaml file and resolve references.

        Args:
            path (str): The path to the yaml file.

        Returns:
            Dict[str, Any]: The loaded data.
        """
        with open(path) as f:
            data = yaml.load(f)
            _resolve_object_paths_recursively(path, data, ObjectPath(), data)
        return data

    @classmethod
    def from_yaml_files(cls, paths: List[str]) -> Self:
        """
        Create a new instance of the model by merging the data from multiple yaml files.

        The data from the files must be either the base model type that follows the schema
        defined in the protobufs, or it must be an overlay that can be merged with the same
        base model type.

        Args:
            paths: The paths to the yaml files.

        Returns:
            The new instance of the model.
        """
        data: Dict[str, Any] = {}
        for path in paths:
            new_data = cls.load_data_with_references(path)

            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=DeprecationWarning, module='jsonmerge')
                data = jsonmerge.merge(data, new_data)
        return cls.model_validate(data)

    def merge(self, other: BaseModel) -> Self:
        """
        Merge this model with another model.

        Args:
            other: The other model to merge with.

        Returns:
            The merged model.
        """
        return self.merge_json(other.model_dump(mode='json', exclude_none=True, exclude_defaults=True))

    def merge_json(self, other: Dict[str, Any]) -> Self:
        """
        Merges a json representation.

        Args:
            other: The other model to merge with as a jsonified dict

        Returns:
            The merged model.
        """
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning, module='jsonmerge')
            data = jsonmerge.merge(self.model_dump(mode='json'), other)
        return self.model_validate(data)

    def apply_cli_args(self, other: BaseModel) -> Self:
        """
        Merge CLI args base model.

        Args:
            other: The other model to merge with.

        Returns:
            The merged model.
        """
        return apply_args(self, other)


def _resolve_object_paths_recursively(root_path: str, root: Dict[str, Any], path_to_ref: ObjectPath, data: Any) -> None:
    """
    Recursively resolves object paths in the data.

    This function traverses the data structure, looking for string values that represent
    object paths. When it finds such a value, it resolves the path and updates the data
    structure accordingly. This process continues recursively for nested structures.

    Args:
        root_path: The path to the root of the data structure.
        root: The root of the data structure.
        path_to_ref: The path to the reference object.
        data: The data to resolve.
    """
    if isinstance(data, dict):
        for field_name, value in data.items():
            _resolve_object_paths_recursively(root_path, root, path_to_ref.add_step(field_name), value)
        return
    elif isinstance(data, list):
        for index, item in enumerate(data):  # type: ignore
            _resolve_object_paths_recursively(root_path, root, path_to_ref.add_index_step(index=index), item)
        return
    elif isinstance(data, str) and EXTERNAL_OBJECT_PATH_REGEX.match(data):
        resolved_value = _resolve_external_ref(root_path, data)
    elif isinstance(data, str) and INTERNAL_OBJECT_PATH_REGEX.match(data):
        path_at_ref = ObjectPath.from_string(data)
        resolved_value = resolve_inner_ref(root, path_to_ref, path_at_ref)
    else:
        return

    steps = path_to_ref.steps.copy()
    container = root
    while len(steps) > 1:
        step = steps.pop(0)
        container = container[step.name]
        if step.index is not None:
            container = container[step.index]

    if steps[0].index is not None:
        container[steps[0].name][steps[0].index] = resolved_value
    else:
        container[steps[0].name] = resolved_value

    _resolve_object_paths_recursively(root_path, root, path_to_ref, resolved_value)


def _resolve_external_ref(root_path: str, path: str) -> Any:
    """
    Resolves an external reference.

    Args:
        path: The path to the external reference.

    Returns:
        The external reference.
    """
    re_match = EXTERNAL_OBJECT_PATH_REGEX.match(path)
    if re_match is None:
        raise ValueError('Expected a path that matches external reference regex')
    file_path = re_match.group(1)

    if not file_path.startswith('/'):
        file_path = str(pathlib.Path(root_path).parent / file_path)

    data = YamlBaseModel.load_data_with_references(file_path)

    if re_match.group(2) is None:
        return data

    path_at_ref = ObjectPath.from_string(re_match.group(2))
    return resolve_inner_ref(data, ObjectPath(), path_at_ref)



import pgml_out.cli_demo_base as base
import pgml_out.cli_demo_sweep as sweep
import pgml_out.cli_demo_cli_args as cli_args
import typer
import py_gen_ml as pgml
import optuna
import typing

app = typer.Typer(pretty_exceptions_enable=False)

def run_trial(
    cli_demo: base.CLIDemo,
    trial: typing.Optional[optuna.Trial] = None
) -> typing.Union[float, typing.Sequence[float]]:
    """
    Run a trial with the given values for cli_demo. The sampled hyperparameters have
    already been added to the trial.
    """
    # TODO: Implement this function
    return 0.0

@pgml.pgml_cmd(app=app)
def main(
    config_paths: typing.List[str] = typer.Option(..., help="Paths to config files"),
    sweep_paths: typing.List[str] = typer.Option(
        default_factory=list,
        help="Paths to sweep files"
    ),
    cli_args: cli_args.CLIDemoArgs = typer.Option(...),
) -> None:
    cli_demo = base.CLIDemo.from_yaml_files(config_paths)
    cli_demo = cli_demo.apply_cli_args(cli_args)
    if len(sweep_paths) == 0:
        run_trial(cli_demo)
        return
    cli_demo_sweep = sweep.CLIDemoSweep.from_yaml_files(sweep_paths)

    def objective(trial: optuna.Trial) -> typing.Union[
        float,
        typing.Sequence[float]
    ]:
        optuna_sampler = pgml.OptunaSampler(trial)
        cli_demo_patch = optuna_sampler.sample(cli_demo_sweep)
        cli_demo_patched = cli_demo.merge(cli_demo_patch)
        objective_value = run_trial(cli_demo_patched, trial)
        return objective_value

    study = optuna.create_study(direction='maximize')
    study.optimize(objective, n_trials=100)


if __name__ == "__main__":
    app()


import typing

import optuna
import pgml_out_test.unit_base as base
import pgml_out_test.unit_cli_args as cli_args
import pgml_out_test.unit_sweep as sweep
import typer

import py_gen_ml as pgml

app = typer.Typer(pretty_exceptions_enable=False)


def run_trial(
    explicit_cli_arg_test: base.ExplicitCLIArgTest,
    trial: typing.Optional[optuna.Trial] = None,
) -> typing.Union[float, typing.Sequence[float]]:
    """
    Run a trial with the given values for explicit_cli_arg_test. The sampled
    hyperparameters have already been added to the trial.
    """
    # TODO: Implement this function
    return 0.0


@pgml.pgml_cmd(app=app)
def main(
    config_paths: typing.List[str] = typer.Option(..., help='Paths to config files'),
    sweep_paths: typing.List[str] = typer.Option(
        default_factory=list,
        help='Paths to sweep files',
    ),
    cli_args: cli_args.ExplicitCLIArgTestArgs = typer.Option(...),
) -> None:
    explicit_cli_arg_test = base.ExplicitCLIArgTest.from_yaml_files(config_paths)
    explicit_cli_arg_test = explicit_cli_arg_test.apply_cli_args(cli_args)
    if len(sweep_paths) == 0:
        run_trial(explicit_cli_arg_test)
        return
    explicit_cli_arg_test_sweep = sweep.ExplicitCLIArgTestSweep.from_yaml_files(sweep_paths)

    def objective(trial: optuna.Trial) -> typing.Union[
        float,
        typing.Sequence[float],
    ]:
        optuna_sampler = pgml.OptunaSampler(trial)
        explicit_cli_arg_test_patch = optuna_sampler.sample(explicit_cli_arg_test_sweep)
        explicit_cli_arg_test_patched = explicit_cli_arg_test.merge(explicit_cli_arg_test_patch)
        objective_value = run_trial(explicit_cli_arg_test_patched, trial)
        return objective_value

    study = optuna.create_study(direction='maximize')
    study.optimize(objective, n_trials=100)


if __name__ == '__main__':
    app()



# Autogenerated code. DO NOT EDIT.
import py_gen_ml as pgml
import typing

import pydantic
import typer

from . import config_base as base


class ProjectArgs(pgml.YamlBaseModel):
    """Global configuration"""

    conv_activation: typing.Annotated[
        typing.Optional[base.Activation],
        typer.
        Option(help="Activation function. Maps to 'net.conv_net.block.activation'"),
        pydantic.Field(None),
        pgml.ArgRef("net.conv_net.block.activation"),
    ]
    """Activation function"""

    head_activation: typing.Annotated[
        typing.Optional[base.Activation],
        typer.Option(help="Activation function. Maps to 'net.head.block.activation'"),
        pydantic.Field(None),
        pgml.ArgRef("net.head.block.activation"),
    ]
    """Activation function"""

    num_conv_layers: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of layers. Maps to 'net.conv_net.num_layers'"),
        pydantic.Field(None),
        pgml.ArgRef("net.conv_net.num_layers"),
    ]
    """Number of layers"""

    num_mlp_layers: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of layers. Maps to 'net.head.num_layers'"),
        pydantic.Field(None),
        pgml.ArgRef("net.head.num_layers"),
    ]
    """Number of layers"""

    learning_rate: typing.Annotated[
        typing.Optional[float],
        typer.Option(help="Learning rate. Maps to 'learning_rate'"),
        pydantic.Field(None),
        pgml.ArgRef("learning_rate"),
    ]
    """Learning rate"""

    beta1: typing.Annotated[
        typing.Optional[float],
        typer.Option(help="Decay rate. Maps to 'beta1'"),
        pydantic.Field(None),
        pgml.ArgRef("beta1"),
    ]
    """Decay rate"""

    batch_size: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Batch size for a single GPU. Maps to 'batch_size'"),
        pydantic.Field(None),
        pgml.ArgRef("batch_size"),
    ]
    """Batch size for a single GPU"""

    num_epochs: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of epochs to train. Maps to 'num_epochs'"),
        pydantic.Field(None),
        pgml.ArgRef("num_epochs"),
    ]
    """Number of epochs to train"""

    out_channels: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of output channels. Maps to 'out_channels'"),
        pydantic.Field(None),
        pgml.ArgRef("out_channels"),
    ]
    """Number of output channels"""

    kernel_size: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Square kernel size. Maps to 'kernel_size'"),
        pydantic.Field(None),
        pgml.ArgRef("kernel_size"),
    ]
    """Square kernel size"""

    pool_size: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Square pool size. Maps to 'pool_size'"),
        pydantic.Field(None),
        pgml.ArgRef("pool_size"),
    ]
    """Square pool size"""

    out_features: typing.Annotated[
        typing.Optional[int],
        typer.Option(help="Number of output features. Maps to 'out_features'"),
        pydantic.Field(None),
        pgml.ArgRef("out_features"),
    ]
    """Number of output features"""



import typing

import py_gen_ml as pgml

from . import enum_demo_patch as patch
from . import enum_demo_base as base

ActivationSweepField = typing.Union[
    pgml.Choice[base.Activation],
    typing.Literal['any'],
    base.Activation,
]


class MLPSweep(pgml.Sweeper[patch.MLPPatch]):
    """MLP configuration"""

    activation: typing.Optional[ActivationSweepField] = None
    """Activation function"""

    num_layers: typing.Optional[pgml.IntSweep] = None
    """Number of layers"""



MLPSweepField = typing.Union[
    MLPSweep,
    pgml.NestedChoice[MLPSweep, patch.MLPPatch],  # type: ignore
]

